"""
Base views for enterprise data api v1.
"""
from edx_rbac.mixins import PermissionRequiredMixin
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from edx_rest_framework_extensions.paginators import DefaultPagination

from enterprise_data.constants import ANALYTICS_API_VERSION_1


class EnterpriseViewSetMixin(PermissionRequiredMixin):
    """
    Base class for all Enterprise view sets.
    """
    authentication_classes = (JwtAuthentication,)
    pagination_class = DefaultPagination
    permission_required = 'can_access_enterprise'
    API_VERSION = ANALYTICS_API_VERSION_1

    def paginate_queryset(self, queryset):
        """
        Allows no_page query param to skip pagination
        """
        if 'no_page' in self.request.query_params:
            return None
        return super().paginate_queryset(queryset)



"""
Test cases for enterprise_admin views
"""
from unittest import mock

import ddt
from mock import patch
from pytest import mark
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITransactionTestCase

from enterprise_data.tests.mixins import JWTTestMixin
from enterprise_data.tests.test_utils import (
    UserFactory,
    get_dummy_engagements_data,
    get_dummy_enrollments_data,
    get_dummy_enterprise_api_data,
)
from enterprise_data_roles.constants import ENTERPRISE_DATA_ADMIN_ROLE
from enterprise_data_roles.models import EnterpriseDataFeatureRole, EnterpriseDataRoleAssignment


@ddt.ddt
@mark.django_db
class TestEnterpriseAdminAnalyticsAggregatesView(JWTTestMixin, APITransactionTestCase):
    """
    Tests for EnterpriseAdminAnalyticsAggregatesView.
    """

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(name=ENTERPRISE_DATA_ADMIN_ROLE)
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role,
            user=self.user
        )
        self.client.force_authenticate(user=self.user)

        mocked_get_enterprise_customer = mock.patch(
            'enterprise_data.filters.EnterpriseApiClient.get_enterprise_customer',
            return_value=get_dummy_enterprise_api_data()
        )

        self.mocked_get_enterprise_customer = mocked_get_enterprise_customer.start()
        self.addCleanup(mocked_get_enterprise_customer.stop)
        self.enterprise_id = 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c'
        self.set_jwt_cookie()

    def _mock_run_query(self, query):
        """
        mock implementation of run_query.
        """
        if 'fact_enrollment_admin_dash' in query:
            return [
                list(item.values()) for item in get_dummy_enrollments_data(self.enterprise_id, 15)
            ]
        else:
            return [
                list(item.values()) for item in get_dummy_engagements_data(self.enterprise_id, 15)
            ]

    def test_get_admin_analytics_aggregates(self):
        """
        Test to get admin analytics aggregates.
        """
        url = reverse('v1:enterprise-admin-analytics-aggregates', kwargs={'enterprise_id': self.enterprise_id})
        with patch('enterprise_data.admin_analytics.data_loaders.run_query', side_effect=self._mock_run_query):
            response = self.client.get(url)
            assert response.status_code == status.HTTP_200_OK
            assert 'enrolls' in response.json()
            assert 'courses' in response.json()
            assert 'completions' in response.json()
            assert 'hours' in response.json()
            assert 'sessions' in response.json()
            assert 'last_updated_at' in response.json()
            assert 'min_enrollment_date' in response.json()
            assert 'max_enrollment_date' in response.json()



"""
Test the utility functions in the admin_analytics app.
"""
from datetime import datetime, timedelta

from mock import patch

from django.test import TestCase

from enterprise_data.admin_analytics.utils import (
    fetch_and_cache_engagements_data,
    fetch_and_cache_enrollments_data,
    get_cache_timeout,
)


class TestUtils(TestCase):
    """
    Test suite for the utility functions in the admin_analytics package.
    """

    def test_get_cache_timeout(self):
        """
        Validate the get_cache_timeout function.
        """
        now = datetime.now().replace(microsecond=0)
        with patch('enterprise_data.admin_analytics.utils.datetime') as mock_datetime:
            mock_datetime.now.return_value = now
            cache_expiry = now
            self.assertEqual(get_cache_timeout(cache_expiry), 0)

            cache_expiry = now + timedelta(seconds=10)
            self.assertEqual(get_cache_timeout(cache_expiry), 10)

            cache_expiry = now + timedelta(seconds=100)
            self.assertEqual(get_cache_timeout(cache_expiry), 100)

            # Validate the case where cache_expiry is in the past.
            cache_expiry = now - timedelta(seconds=10)
            self.assertEqual(get_cache_timeout(cache_expiry), 0)

    def test_fetch_and_cache_enrollments_data(self):
        """
        Validate the fetch_and_cache_enrollments_data function.
        """
        with patch('enterprise_data.admin_analytics.utils.fetch_enrollment_data') as mock_fetch_enrollment_data:
            with patch('enterprise_data.admin_analytics.utils.TieredCache') as mock_tiered_cache:
                # Simulate the scenario where the data is not found in the cache.
                mock_tiered_cache.get_cached_response.return_value.is_found = False
                mock_fetch_enrollment_data.return_value = 'enrollments'

                enrollments = fetch_and_cache_enrollments_data('enterprise_id', datetime.now() + timedelta(seconds=10))
                self.assertEqual(enrollments, 'enrollments')
                self.assertEqual(mock_tiered_cache.get_cached_response.call_count, 1)
                self.assertEqual(mock_tiered_cache.set_all_tiers.call_count, 1)

    def test_fetch_and_cache_enrollments_data_with_data_cache_found(self):
        """
        Validate the fetch_and_cache_enrollments_data function.
        """
        with patch('enterprise_data.admin_analytics.utils.fetch_enrollment_data') as mock_fetch_enrollment_data:
            with patch('enterprise_data.admin_analytics.utils.TieredCache') as mock_tiered_cache:
                # Simulate the scenario where the data is found in the cache.
                mock_tiered_cache.get_cached_response.return_value.is_found = True
                mock_tiered_cache.get_cached_response.return_value.value = 'cached-enrollments'
                mock_fetch_enrollment_data.return_value = 'enrollments'

                enrollments = fetch_and_cache_enrollments_data('enterprise_id', datetime.now() + timedelta(seconds=10))
                self.assertEqual(enrollments, 'cached-enrollments')
                self.assertEqual(mock_tiered_cache.get_cached_response.call_count, 1)
                self.assertEqual(mock_tiered_cache.set_all_tiers.call_count, 0)

    def test_fetch_and_cache_engagements_data(self):
        """
        Validate the fetch_and_cache_engagements_data function.
        """
        with patch('enterprise_data.admin_analytics.utils.fetch_engagement_data') as mock_fetch_engagement_data:
            with patch('enterprise_data.admin_analytics.utils.TieredCache') as mock_tiered_cache:
                # Simulate the scenario where the data is not found in the cache.
                mock_tiered_cache.get_cached_response.return_value.is_found = False
                mock_fetch_engagement_data.return_value = 'engagements'

                enrollments = fetch_and_cache_engagements_data('enterprise_id', datetime.now() + timedelta(seconds=10))
                self.assertEqual(enrollments, 'engagements')
                self.assertEqual(mock_tiered_cache.get_cached_response.call_count, 1)
                self.assertEqual(mock_tiered_cache.set_all_tiers.call_count, 1)

    def test_fetch_and_cache_engagements_data_with_data_cache_found(self):
        """
        Validate the fetch_and_cache_engagements_data function.
        """
        with patch('enterprise_data.admin_analytics.utils.fetch_engagement_data') as mock_fetch_engagement_data:
            with patch('enterprise_data.admin_analytics.utils.TieredCache') as mock_tiered_cache:
                # Simulate the scenario where the data is found in the cache.
                mock_tiered_cache.get_cached_response.return_value.is_found = True
                mock_tiered_cache.get_cached_response.return_value.value = 'cached-engagements'
                mock_fetch_engagement_data.return_value = 'engagements'

                enrollments = fetch_and_cache_engagements_data('enterprise_id', datetime.now() + timedelta(seconds=10))
                self.assertEqual(enrollments, 'cached-engagements')
                self.assertEqual(mock_tiered_cache.get_cached_response.call_count, 1)
                self.assertEqual(mock_tiered_cache.set_all_tiers.call_count, 0)



"""
Utility functions for fetching data from the database.
"""
from datetime import datetime

from edx_django_utils.cache import TieredCache, get_cache_key

from enterprise_data.admin_analytics.data_loaders import fetch_engagement_data, fetch_enrollment_data


def get_cache_timeout(cache_expiry):
    """
    Helper method to calculate cache timeout in seconds.

    Arguments:
        cache_expiry (datetime): Datetime object denoting the cache expiry.

    Returns:
        (int): Cache timeout in seconds.
    """
    now = datetime.now()
    cache_timeout = 0
    if cache_expiry > now:
        # Calculate cache expiry in seconds from now.
        cache_timeout = (cache_expiry - now).seconds

    return cache_timeout


def fetch_and_cache_enrollments_data(enterprise_id, cache_expiry):
    """
    Helper method to fetch and cache enrollments data.

    Arguments:
        enterprise_id (str): UUID of the enterprise customer in string format.
        cache_expiry (datetime): Datetime object denoting the cache expiry.

    Returns:
        (pandas.DataFrame): The enrollments data.
    """
    cache_key = get_cache_key(
        resource='enterprise-admin-analytics-aggregates-enrollments',
        enterprise_customer=enterprise_id,
    )
    cached_response = TieredCache.get_cached_response(cache_key)

    if cached_response.is_found:
        return cached_response.value
    else:
        enrollments = fetch_enrollment_data(enterprise_id)
        TieredCache.set_all_tiers(
            cache_key, enrollments, get_cache_timeout(cache_expiry)
        )
        return enrollments


def fetch_and_cache_engagements_data(enterprise_id, cache_expiry):
    """
    Helper method to fetch and cache engagements data.

    Arguments:
        enterprise_id (str): UUID of the enterprise customer in string format.
        cache_expiry (datetime): Datetime object denoting the cache expiry.

    Returns:
        (pandas.DataFrame): The engagements data.
    """
    cache_key = get_cache_key(
        resource='enterprise-admin-analytics-aggregates-engagements',
        enterprise_customer=enterprise_id,
    )
    cached_response = TieredCache.get_cached_response(cache_key)

    if cached_response.is_found:
        return cached_response.value
    else:
        engagements = fetch_engagement_data(enterprise_id)
        TieredCache.set_all_tiers(
            cache_key, engagements, get_cache_timeout(cache_expiry)
        )
        return engagements



"""
Utility functions for fetching data from the database.
"""
import numpy
import pandas

from django.http import Http404

from enterprise_data.admin_analytics.database import run_query


def get_select_query(table: str, columns: list, enterprise_uuid: str) -> str:
    """
    Generate a SELECT query for the given table and columns.

    Arguments:
        table (str): The table to query.
        columns (list): The columns to select.
        enterprise_uuid (str): The UUID of the enterprise customer.

    Returns:
        (str): The SELECT query.
    """
    return f'SELECT {", ".join(columns)} FROM {table} WHERE enterprise_customer_uuid = "{enterprise_uuid}"'


def fetch_enrollment_data(enterprise_uuid: str):
    """
    Fetch enrollment data from the database for the given enterprise customer.

    Arguments:
        enterprise_uuid (str): The UUID of the enterprise customer.

    Returns:
        (pandas.DataFrame): The enrollment data.
    """
    enterprise_uuid = enterprise_uuid.replace('-', '')

    columns = [
        'enterprise_customer_name',
        'enterprise_customer_uuid',
        'lms_enrollment_id',
        'user_id',
        'email',
        'course_key',
        'courserun_key',
        'course_id',
        'course_subject',
        'course_title',
        'enterprise_enrollment_date',
        'lms_enrollment_mode',
        'enroll_type',
        'program_title',
        'date_certificate_awarded',
        'grade_percent',
        'cert_awarded',
        'date_certificate_created_raw',
        'passed_date_raw',
        'passed_date',
        'has_passed',
    ]
    query = get_select_query(
        table='fact_enrollment_admin_dash',
        columns=columns,
        enterprise_uuid=enterprise_uuid,
    )

    results = run_query(query=query)
    if not results:
        raise Http404(f'No enrollment data found for enterprise {enterprise_uuid}')

    enrollments = pandas.DataFrame(numpy.array(results), columns=columns)

    # Convert date columns to datetime.
    enrollments['enterprise_enrollment_date'] = enrollments['enterprise_enrollment_date'].astype('datetime64[ns]')
    enrollments['date_certificate_awarded'] = enrollments['date_certificate_awarded'].astype('datetime64[ns]')
    enrollments['date_certificate_created_raw'] = enrollments['date_certificate_created_raw'].astype('datetime64[ns]')
    enrollments['passed_date_raw'] = enrollments['passed_date_raw'].astype('datetime64[ns]')
    enrollments['passed_date'] = enrollments['passed_date'].astype('datetime64[ns]')

    return enrollments


def fetch_engagement_data(enterprise_uuid: str):
    """
    Fetch engagement data from the database for the given enterprise customer.

    Arguments:
        enterprise_uuid (str): The UUID of the enterprise customer.

    Returns:
        (pandas.DataFrame): The engagement data.
    """
    enterprise_uuid = enterprise_uuid.replace('-', '')

    columns = [
        'user_id',
        'email',
        'enterprise_customer_uuid',
        'course_key',
        'enroll_type',
        'activity_date',
        'course_title',
        'course_subject',
        'is_engaged',
        'is_engaged_video',
        'is_engaged_forum',
        'is_engaged_problem',
        'is_active',
        'learning_time_seconds',
    ]
    query = get_select_query(
        table='fact_enrollment_engagement_day_admin_dash', columns=columns, enterprise_uuid=enterprise_uuid
    )

    results = run_query(query=query)
    if not results:
        raise Http404(f'No engagement data found for enterprise {enterprise_uuid}')

    engagement = pandas.DataFrame(numpy.array(results), columns=columns)
    engagement['activity_date'] = engagement['activity_date'].astype('datetime64[ns]')

    return engagement


def fetch_max_enrollment_datetime():
    """
    Fetch the latest created date from the enterprise_learner_enrollment table.

    created will be same for all records as this is added at the time of data load. Which is when the async process
    populates the data in the table. We can use this to get the latest data load time.
    """
    query = "SELECT MAX(created) FROM enterprise_learner_enrollment"
    results = run_query(query)
    if not results:
        return None
    return pandas.to_datetime(results[0][0])



"""
Test the utility functions in the admin_analytics app for data loading operations.
"""
from uuid import uuid4

import pytest
from mock import patch

from django.http import Http404
from django.test import TestCase

from enterprise_data.admin_analytics.data_loaders import (
    fetch_engagement_data,
    fetch_enrollment_data,
    fetch_max_enrollment_datetime,
)
from enterprise_data.tests.test_utils import get_dummy_engagements_data, get_dummy_enrollments_data


class TestDataLoaders(TestCase):
    """
    Test suite for the utility functions in the admin_analytics package for data loading operations.
    """

    def test_fetch_max_enrollment_datetime(self):
        """
        Validate the fetch_max_enrollment_datetime function.
        """
        with patch('enterprise_data.admin_analytics.data_loaders.run_query') as mock_run_query:
            mock_run_query.return_value = [['2024-07-26']]

            max_enrollment_date = fetch_max_enrollment_datetime()
            self.assertEqual(max_enrollment_date.strftime('%Y-%m-%d'), '2024-07-26')

            # Validate the case where the query returns an empty result.
            mock_run_query.return_value = []
            max_enrollment_date = fetch_max_enrollment_datetime()
            self.assertIsNone(max_enrollment_date)

    def test_fetch_engagement_data(self):
        """
        Validate the fetch_engagement_data function.
        """
        with patch('enterprise_data.admin_analytics.data_loaders.run_query') as mock_run_query:
            enterprise_uuid = str(uuid4())
            mock_run_query.return_value = [
                list(item.values()) for item in get_dummy_engagements_data(enterprise_uuid, 10)
            ]

            engagement_data = fetch_engagement_data(enterprise_uuid)
            self.assertEqual(engagement_data.shape, (10, 14))

    def test_fetch_engagement_data_empty_data(self):
        """
        Validate the fetch_engagement_data function behavior when no data is returned from the query.
        """
        with patch('enterprise_data.admin_analytics.data_loaders.run_query') as mock_run_query:
            mock_run_query.return_value = []
            enterprise_uuid = str(uuid4())
            with pytest.raises(Http404) as error:
                fetch_engagement_data(enterprise_uuid)
            error.value.message = f'No engagement data found for enterprise {enterprise_uuid}'

    def test_fetch_enrollment_data(self):
        """
        Validate the fetch_enrollment_data function.
        """
        with patch('enterprise_data.admin_analytics.data_loaders.run_query') as mock_run_query:
            enterprise_uuid = str(uuid4())
            mock_run_query.return_value = [
                list(item.values()) for item in get_dummy_enrollments_data(enterprise_uuid)
            ]

            enrollment_data = fetch_enrollment_data(enterprise_uuid)
            self.assertEqual(enrollment_data.shape, (10, 21))

    def test_fetch_enrollment_data_empty_data(self):
        """
        Validate the fetch_enrollment_data function behavior when no data is returned from the query.
        """
        with patch('enterprise_data.admin_analytics.data_loaders.run_query') as mock_run_query:
            mock_run_query.return_value = []
            enterprise_uuid = str(uuid4())
            with pytest.raises(Http404) as error:
                fetch_enrollment_data(enterprise_uuid)
            error.value.message = f'No enrollment data found for enterprise {enterprise_uuid}'



"""
Views for enterprise admin api v1.
"""
from datetime import datetime, timedelta

from edx_rbac.decorators import permission_required
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK, HTTP_404_NOT_FOUND
from rest_framework.views import APIView

from enterprise_data.admin_analytics.data_loaders import fetch_max_enrollment_datetime
from enterprise_data.admin_analytics.utils import fetch_and_cache_engagements_data, fetch_and_cache_enrollments_data
from enterprise_data.api.v1 import serializers
from enterprise_data.models import EnterpriseAdminLearnerProgress, EnterpriseAdminSummarizeInsights
from enterprise_data.utils import date_filter


class EnterpriseAdminInsightsView(APIView):
    """
    API for getting the enterprise admin insights.
    """
    authentication_classes = (JwtAuthentication,)
    http_method_names = ['get']

    @permission_required('can_access_enterprise', fn=lambda request, enterprise_id: enterprise_id)
    def get(self, request, enterprise_id):
        """
        HTTP GET endpoint to retrieve the enterprise admin insights
        """
        response_data = {}
        learner_progress = {}
        learner_engagement = {}

        try:
            learner_progress = EnterpriseAdminLearnerProgress.objects.get(enterprise_customer_uuid=enterprise_id)
            learner_progress = serializers.EnterpriseAdminLearnerProgressSerializer(learner_progress).data
            response_data['learner_progress'] = learner_progress
        except EnterpriseAdminLearnerProgress.DoesNotExist:
            pass

        try:
            learner_engagement = EnterpriseAdminSummarizeInsights.objects.get(enterprise_customer_uuid=enterprise_id)
            learner_engagement = serializers.EnterpriseAdminSummarizeInsightsSerializer(learner_engagement).data
            response_data['learner_engagement'] = learner_engagement
        except EnterpriseAdminSummarizeInsights.DoesNotExist:
            pass

        status = HTTP_200_OK
        if learner_progress == {} and learner_engagement == {}:
            status = HTTP_404_NOT_FOUND

        return Response(data=response_data, status=status)


class EnterpriseAdminAnalyticsAggregatesView(APIView):
    """
    API for getting the enterprise admin analytics aggregates.
    """
    authentication_classes = (JwtAuthentication,)
    http_method_names = ['get']

    @permission_required('can_access_enterprise', fn=lambda request, enterprise_id: enterprise_id)
    def get(self, request, enterprise_id):
        """
        HTTP GET endpoint to retrieve the enterprise admin aggregate data.
        """
        serializer = serializers.AdminAnalyticsAggregatesQueryParamsSerializer(data=request.GET)
        serializer.is_valid(raise_exception=True)

        last_updated_at = fetch_max_enrollment_datetime()
        cache_expiry = last_updated_at + timedelta(days=1) if last_updated_at else datetime.now()

        enrollment = fetch_and_cache_enrollments_data(enterprise_id, cache_expiry).copy()
        engagement = fetch_and_cache_engagements_data(enterprise_id, cache_expiry).copy()
        # Use start and end date if provided by the client, if client has not provided then use
        # 1. minimum enrollment date from the data as the start_date
        # 2. today's date as the end_date
        start_date = serializer.data.get('start_date', enrollment.enterprise_enrollment_date.min())
        end_date = serializer.data.get('end_date', datetime.now())

        # Date filtering.
        dff = date_filter(
            start=start_date, end=end_date, data_frame=enrollment.copy(), date_column='enterprise_enrollment_date'
        )

        enrolls = len(dff)
        courses = len(dff.course_key.unique())

        dff = date_filter(start=start_date, end=end_date, data_frame=enrollment.copy(), date_column='passed_date')

        completions = dff.has_passed.sum()

        # Date filtering.
        dff = date_filter(start=start_date, end=end_date, data_frame=engagement.copy(), date_column='activity_date')

        hours = round(dff.learning_time_seconds.sum() / 60 / 60, 1)
        sessions = dff.is_engaged.sum()

        return Response(data={
            'enrolls': enrolls,
            'courses': courses,
            'completions': completions,
            'hours': hours,
            'sessions': sessions,
            'last_updated_at': last_updated_at.date() if last_updated_at else None,
            'min_enrollment_date': enrollment.enterprise_enrollment_date.min().date(),
            'max_enrollment_date': enrollment.enterprise_enrollment_date.max().date(),
        }, status=HTTP_200_OK)
